# Django library.
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

def retrieve(name):
    try:
        return getattr(settings, name, False)
    except ImproperlyConfigured:
        # To handle the auto-generation of documentations.
        return False


IMAGEIT_MAX_UPLOAD_SIZE_MB = retrieve('IMAGEIT_MAX_UPLOAD_SIZE_MB') or 5
IMAGEIT_MAX_SAVE_SIZE_MB = retrieve('IMAGEIT_MAX_SAVE_SIZE_MB') or 5
IMAGEIT_ACCEPTED_CONTENT_TYPES = retrieve('IMAGEIT_ACCEPTED_CONTENT_TYPES') or 'image/jpeg', 'image/png', 'image/svg+xml'
IMAGEIT_SVG_CONTENT_TYPE = retrieve('IMAGEIT_SVG_CONTENT_TYPE') or 'image/svg+xml'
IMAGEIT_DPI = retrieve('IMAGEIT_DPI') or 90
IMAGEIT_DEFAULT_IMAGE_PROPS = retrieve('IMAGEIT_DEFAULT_IMAGE_PROPS') or {"max_width": 1000, "max_height": 1000, "quality": IMAGEIT_DPI, "upscale": False}
IMAGEIT_IMAGE_PROPERTIES = retrieve('IMAGEIT_IMAGE_PROPERTIES') or ['max_width', 'max_height', 'quality', 'upscale', 'max_save_size_mb']
IMAGEIT_CROP_PROPERTIES = retrieve('IMAGEIT_CROP_PROPERTIES') or ["x1", "y1", "x2", "y2"]


from django.contrib.admin.widgets import AdminFileWidget
from django.core.validators import RegexValidator
from django.forms import MultiValueField, FileField, FloatField
from django.forms.widgets import HiddenInput

from .widgets import ScaleItImageWidget, CropItImageWidget


class ScaleItImageFormField(FileField):
    # Change widget to custom one
    widget = ScaleItImageWidget

    def __init__(self, widget=None, max_length=None, allow_empty_file=False, *args, **kwargs):
        # Stop Django admin from overriding widget with default AdminFileWidget
        # Occurs when instantiating widgets from admin views
        # See : https://github.com/django/django/blob/master/django/contrib/admin/options.py
        super(ScaleItImageFormField, self).__init__(*args, **kwargs)
        if isinstance(widget, AdminFileWidget):
            self.widget = ScaleItImageWidget

    def clean(self, data, initial=None):
        # If you want image in cleaned_data to be scaled/cropped
        # It must be completed here as model.clean is completed in form.post_clean
        # Img props will need to be passed from model.form_class
            # Should probably validate crop inputs here
        return super().clean(data, initial)

class CropItImageFormField(MultiValueField):
    widget = CropItImageWidget(widgets=[ScaleItImageWidget, HiddenInput, HiddenInput, HiddenInput, HiddenInput,])

    def __init__(self, max_length=None, widget=None, **kwargs):
        # Define one message for all fields.
        error_messages = {
            'incomplete': 'Missing values required for image crop (required: x, y, width, height).',
        }
        # Or define a different message for each field.
        fields = (
            ScaleItImageFormField(
                error_messages={'incomplete': 'Please select an image to upload.'},
                required=kwargs.get('required'),
            ),
            FloatField(
                error_messages={'incomplete': 'Please eneter a starting x value for image crop.'},
                validators=[RegexValidator(r'^[0-9]+.[0-9]+$', 'Please eneter an x1 value for image crop.')],
                required=True,
            ),
            FloatField(
                validators=[RegexValidator(r'^[0-9]+.[0-9]+$', 'Please eneter a y1 value for image crop.')],
                required=True,
            ),
            FloatField(
                validators=[RegexValidator(r'^[0-9]+.[0-9]+$', 'Please eneter an x2 value for image crop.')],
                required=True,
            ),
            FloatField(
                validators=[RegexValidator(r'^[0-9]+.[0-9]+$', 'Please eneter a y2 value for image crop.')],
                required=True,
            ),
        )
        super(CropItImageFormField, self).__init__(
            error_messages=error_messages, fields=fields,
            require_all_fields=False, **kwargs
        )


    # Compresses inputs into single file object with crop values as properties
    # data_list: list of all cleaned input values
    def compress(self, data_list):
        img = data_list[0]
        if img is not None and img is not False:
            img.crop_props = {"x1": data_list[1], "y1": data_list[2],"x2": data_list[3], "y2": data_list[4]}
        return img


import filetype


class svgType(filetype.Type):
    """
    Implements a mime type matcher for svg files
    """
    MIME = 'image/svg+xml'
    EXTENSION = 'svg'

    def __init__(self):
        super(svgType, self).__init__(
            mime=svgType.MIME,
            extension=svgType.EXTENSION
        )

    def match(self, buf):
        str_buf = str(buf)
        location = str_buf.lower().find('<svg')
        return location >= 0


# Create your models here.
from django.db.models import FileField
from django.utils.translation import gettext_lazy as _


from .fields import ScaleItImageFormField, CropItImageFormField
from .settings import (
    IMAGEIT_DEFAULT_IMAGE_PROPS,
    IMAGEIT_IMAGE_PROPERTIES,
)
from . import utils


class ScaleItImageField(FileField):
    description = _("Image field that scales to the bounds of max_height and max_width")

    def __init__(self, *args, **kwargs):
        # Dictionary containing propeties for cropping and or scaling image
        # max_width, max_height, quality, max_upload_size, crop_props (for crop field)
        self.img_props = dict(IMAGEIT_DEFAULT_IMAGE_PROPS)

        for key, val in list(kwargs.items()):
            if key.lower() in IMAGEIT_IMAGE_PROPERTIES:
                self.img_props[key.lower()] = kwargs.pop(key)
        super(ScaleItImageField, self).__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        # Update form field class used. Can also pass kwargs to form class from defaults
        defaults = {
            'form_class': ScaleItImageFormField,
        }
        defaults.update(kwargs)
        # for passing kwargs to form_class if required
        # defaults.update(**self.img_props)
        # Returns instantiated form_class class
        return super(ScaleItImageField, self).formfield(**defaults)

    # Required to reconstruct state of model field instance. 
    # (Any updated kwargs must be returned if varying from default value used in init)
    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if self.img_props != IMAGEIT_DEFAULT_IMAGE_PROPS:
            kwargs.update(self.img_props)
        return name, path, args, kwargs
    
    def _clean(self, value):
        scaled_img = utils.process_upload(value, self.img_props)
        scaled_img.cleaned = True
        return scaled_img

    # Returns a cleaned django InMemoryUploadedFile instance
    def clean(self, value, model_instance):
        scaled_img = self._clean(value)
        return super().clean(scaled_img, model_instance)

    # Scaling/Cropping and file Checks are verified in model as apposed to form,
    # ensuring files are always validated, even when adding programatically
    def pre_save(self, model_instance, add):
        file = getattr(model_instance, self.attname)

        # Prevents re-running of clean when using from form 
        # as form.post_clean triggers model clean methods
        if file and not hasattr(file, 'cleaned'):
            file.file = self._clean(file.file)
            setattr(model_instance, self.attname, file)
        return super().pre_save(model_instance, add)


class CropItImageField(ScaleItImageField):
    description = _("Image field that crops an image then scales it to the bounds of max_height and max_width")

    def formfield(self, **kwargs):
        defaults = {
            'form_class': CropItImageFormField,
        }
        defaults.update(kwargs)
        return super(CropItImageField, self).formfield(**defaults)

    # value: InMemoryUploadedFile
    def clean(self, value, model_instance):
        # Check to see if crop_props are present in file. 
        # (hasattr checks to avoid throwing attribute error (no file) on unchanged initial file (s3BotoFile))
        if hasattr(value, 'file') and hasattr(value.file, 'crop_props'):
            self.img_props.update({'crop_props': value.file.crop_props})

        return super().clean(value, model_instance)


from django.core.exceptions import ImproperlyConfigured
from django.forms.widgets import MultiWidget, ClearableFileInput
from django.utils.translation import gettext_lazy as _

from .settings import (
    IMAGEIT_MAX_UPLOAD_SIZE_MB
)

class ScaleItImageWidget(ClearableFileInput):
    template_name = 'imageit/widgets/scale_it_widget.html'
    initial_text = 'Current'

    def __init__(self, *args, **kwargs):
        if kwargs.get('multiple') == True:
            raise ImproperlyConfigured(_('Scaleit image widget does not support multi-file selectors'), code="Disallowed widget attrs")
        super(ScaleItImageWidget, self).__init__(*args, **kwargs)

    def get_context(self, name, value, attrs):
        if not attrs.get("data-max_upload_size"):
            attrs["data-max_upload_size"] = IMAGEIT_MAX_UPLOAD_SIZE_MB
        context = super().get_context(name, value, attrs)
        return context

    class Media:
        css = {"all": ('imageit/css/imageit.css',),}
        js = ('imageit/js/imageit.js',)


class CropItImageWidget(MultiWidget):
    template_name = 'imageit/widgets/crop_it_widget.html'

    def __init__(self, *args, **kwargs):
        if kwargs.get('multiple') == True:
            raise ImproperlyConfigured(_('Cropit image widget does not support multi-file selectors'), code="Disallowed widget attrs")
        super(CropItImageWidget, self).__init__(*args, **kwargs)

    def get_context(self, name, value, attrs):
        if not attrs.get("data-max_upload_size"):
            attrs["data-max_upload_size"] = IMAGEIT_MAX_UPLOAD_SIZE_MB
        context = super().get_context(name, value, attrs)
        return context

    def decompress(self, value):
        if value:
            return [value, 0, 0, 0, 0]
        return [None, 0, 0, 0, 0]

    class Media:
        css = {"all": ('imageit/css/imageit.css', 'https://cdnjs.cloudflare.com/ajax/libs/cropperjs/1.5.9/cropper.min.css'),}
        js = ('https://cdnjs.cloudflare.com/ajax/libs/cropperjs/1.5.9/cropper.min.js', '/static/imageit/js/imageit.js')


import sys

from django.core.exceptions import ValidationError
from django.core.files.base import File
from django.utils.translation import gettext_lazy as _
import filetype
from io import BytesIO
from os import SEEK_END
from PIL import Image
from re import search, IGNORECASE, MULTILINE
from xml.etree.ElementTree import fromstring

from .filetypes import svgType
from .settings import (
    IMAGEIT_DEFAULT_IMAGE_PROPS,
    IMAGEIT_ACCEPTED_CONTENT_TYPES,
    IMAGEIT_MAX_UPLOAD_SIZE_MB,
    IMAGEIT_MAX_SAVE_SIZE_MB,
    IMAGEIT_SVG_CONTENT_TYPE
)

# data: Django InMemoryUploadedFile Instance
# img_props : dict {max_width, max_height, upscale, crop_props(opt.):{}}
def process_upload(data, img_props=None):
    # Check to see if a file was uploaded
    if isinstance(data, File):
        # Add svg mime type matcher
        filetype.add_type(svgType())
        data.content_type = filetype.guess(data.read(256)).mime
        if data.content_type is None:
            raise ValidationError(_("Unable to derive the content type of the file submitted"), code='file_invalid')

        # Determine the size of the file
        if hasattr(data, 'size'):
            size = data.size
        if hasattr(data, 'tell') and hasattr(data, 'seek'):
            pos = data.tell()
            data.seek(0, SEEK_END)
            size = data.tell()
            data.seek(pos)
        else:
            raise AttributeError("Unable to determine the file's size.")

        # Check to make sure image is an accepted content type
        if data.content_type in IMAGEIT_ACCEPTED_CONTENT_TYPES:
            # If file size exceeds max allowed, Raise error. Else process Image
            if size > (IMAGEIT_MAX_UPLOAD_SIZE_MB * 1024 * 1024):
                raise ValidationError(_("Uploaded file exceeds maximum allowed size of %(size)sMB.") %{ 'size': IMAGEIT_MAX_UPLOAD_SIZE_MB}, code="file_invalid")
            else:
                # Process vector or raster depending on file type
                if data.content_type == IMAGEIT_SVG_CONTENT_TYPE:
                    data = process_vector(data)
                else:
                    data = process_raster(data, img_props)
                
                # Validate file size after resampling 
                if size > (img_props.get('max_save_size_mb', IMAGEIT_MAX_SAVE_SIZE_MB) * 1024 * 1024):
                    raise ValidationError(_("Uploaded file exceeds maximum allowed size of %(size)sMB.") %{ 'size': IMAGEIT_MAX_SAVE_SIZE_MB}, code="file_invalid")
        else:
            raise ValidationError(_("Unsupported image format. Must be one of %(opts)s") % { 'opts': IMAGEIT_ACCEPTED_CONTENT_TYPES}, code='file_invalid')
    else:
        raise ValidationError(_("Process Upload requires a valid django File instance. %(type)s") % { 'type': type(data)}, code='file_invalid')
    return data


def process_vector(image):
    # Image: Django File Instance
    if contains_javascript(image):
        raise ValidationError(_("File rejected: JavaScript was detected within the file."), code='file_invalid')
    return image


def process_raster(image, img_props=None):
    # Image: Django File Instance
    if img_props == None or len(img_props) == 0:
        img_props = IMAGEIT_DEFAULT_IMAGE_PROPS

    # Returns Django File Instance
    scaled_image = resize(image, img_props)
    return scaled_image


def contains_javascript(image):
    image.file.seek(0)
    file_str = str(image.file.read(), encoding='UTF-8')

    # Filters against "script" / "if (.)" / "for (.)".
    pattern = r'(?i)(<\s*\bscript\b.*>.*?)|(.*\bif\b\s*\(.?.*\))|(.*\bfor\b\s*\(.*\))'

    found = search(
        pattern=pattern,
        string=file_str,
        flags=IGNORECASE | MULTILINE
    )
    
    if found is not None:
        return True

    parsed_xml = (
        (attribute, value)
        for elm in fromstring(file_str).iter()
        for attribute, value in elm.attrib.items()
    )

    for key, val in parsed_xml:
        if '"' in val or "'" in val:
            return True

    return False


def resize(image, img_props):
    # Image: Django File Instance
    # Open image and store format/metadata.
    pil_image = Image.open(image)
    pil_image_format, pil_image_info = pil_image.format, pil_image.info
    if img_props.get('quality'):
        pil_image_info['quality'] = img_props.get('quality')

    # Force PIL to load image data.
    pil_image.load()

    #Retrieve Crop props
    crop_props = img_props.pop('crop_props', False)

    #Check that all required co-oords are present and valid
    if crop_props and all(x in crop_props for x in ("x1", "y1", "x2", "y2")):
        if crop_props.get('x1') - crop_props.get('x2') == 0 or crop_props.get('y1') - crop_props.get('y2') == 0:
            raise ValidationError(_("Cropped image (%(file)s) cannot have a width or height or zero!" % {'file': image.name}), code='crop_invalid')
        else:
            pil_image = crop(pil_image, crop_props)
    elif crop_props and any(x in crop_props for x in ("x1", "y1", "x2", "y2")):
        raise ValidationError(_("Incomplete Cooridinates for cropping were recieved! Requires: x1, y1, x2, y2"), code='missing_crop_values')
    pil_image = scale(pil_image, img_props)

    # Close image and replace format/metadata, as PIL blows this away.
    pil_image.format, pil_image.info = pil_image_format, pil_image_info

    #Save pil image back to bytesio file
    extension = image.content_type.split('/')[-1].upper()
    bytes_image = BytesIO()
    try:
        pil_image.save(bytes_image, extension)
        bytes_image.seek(0, SEEK_END)
    except Exception as e:
        raise ValidationError(_("Error: %(e)s. There may be an issue with your image file.") % { 'e': e })
    pil_image.close()
    image.file = bytes_image
    return image


def crop(image, props):
    # Image: PIL Image
    image_width, image_height = map(float, image.size)

    x1, y1, x2, y2 = (props.get('x1'), props.get('y1'), props.get('x2'), props.get('y2'),)
    # Check if image actually requires cropping
    if not((x1 == 0 and x2 == image_width) or (y1 == 0 and y2 == image_height)):
        box = [
            int(x1),
            int(y1),
            int(x2),
            int(y2)
        ]
        # Crop the image
        image = image.crop(box)
    return image


def scale(image, props):
    #Image: PIL Image

    max_width, max_height, upscale = (props.get('max_width'), props.get('max_height'), props.get('upscale'))
    image_width, image_height = map(float, image.size)

    #Calculate scale ratio ensuring no side is longer than max dimenstions
    ratio = min(max_width / image_width, max_height / image_height)

    # Scale image if image must be scaled down or if upscale is set to true
    if ratio < 1.0 or (ratio > 1.0 and upscale):
        scaled_width = image_width * ratio
        scaled_height = image_height * ratio
    
        image = image.resize(
            (int(scaled_width), int(scaled_height)),
            resample=Image.ANTIALIAS
        )
    return image



from setuptools import setup

with open("README.rst", "r") as fh:
    long_description = fh.read()

setup(
    name='django-imageit',
    version='0.0.3',
    url="https://github.com/scott-j5/django-imageit",
    author="Scott James",
    author_email="scottjames@byitegroup.com",
    description="Image processing plugin built for Django",
    packages=['imageit'],
    long_description=long_description,
    long_description_content_type="text/x-rst",

    install_requires = [
        "django >= 3",
        "pillow > 8.0.0",
        "filetype > 1.0.0",
    ],
    extras_require = {
        "dev": [
            "twine >= 3.4",
        ]
    },
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: JavaScript',
        'Framework :: Django :: 3.0',
        'Framework :: Django :: 3.1',
        'Framework :: Django :: 3.2',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Multimedia :: Graphics',
        'Topic :: Multimedia :: Graphics :: Presentation',
    ],
    keywords='django images imageit image upload scale crop resize',
)